/* ccgo: gowin.cc
 *
 * GoWin GoWinPlay GoWinObserve GoWinMatch class definitions
 * 
 * Copyright 2002 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include "gowin.hh"
#include <sstream>
#include <iostream>

#define OTHER_STONE(s) ((s == B_STONE) ? W_STONE : B_STONE)
#define OTHER_TURN(t) ((t == B_TURN) ? W_TURN : B_TURN)
#define TURN_STONE(t) ((t == B_TURN) ? B_STONE : W_STONE)
#define TURN_OTHER_STONE(t) ((t == B_TURN) ? W_STONE : B_STONE)
#define STONE_OTHER_TURN(s) ((s == B_STONE) ? W_TURN : B_TURN)

GoWin::GoWin(int ss) :
	Gtk::Window(GTK_WINDOW_TOPLEVEL),
	GoGame(ss),
	bd(ss),
	numm("  0"),
	dma(0, 0, 0),
	lastm("start"),
	gt(* this)
{
	dm = 0;
	ms = STAGE_PLAYING;

	// set_title(t);
	set_default_size(400, 440);

	Gtk::VBox * v;
	Gtk::VBox * vv;
	Gtk::HBox * h;
	Gtk::HBox * hh;
	Gtk::SpinButton * s;
	Gtk::Button * b;
	Gtk::AspectFrame * a;
	Gtk::VPaned * p;

	p = manage(new Gtk::VPaned);
	add(* p);

	v = manage(new Gtk::VBox);
	p->pack1(* v, true, true);

	// Top Area
	v->pack_start(top_area, false);
	top_area.show();

	h = manage(new Gtk::HBox);
	v->pack_start(* h);
	a = manage(new Gtk::AspectFrame(0.5, 0.5));
	a->add(bd);
	bd.show();
	h->pack_start(* a);
	a->show();

	vv = manage(new Gtk::VBox);
	h->pack_start(* vv, false);

	Gtk::Label * l = manage(new Gtk::Label("Move"));
	vv->pack_start(* l, false);
	l->show();

	vv->pack_start(numm, false);
	numm.show();

	s = manage(new Gtk::SpinButton(dma, 0, 0));
	vv->pack_start(* s, false);
	dma.value_changed.connect(slot(this, & GoWin::change_display));
	s->show();

	vv->pack_start(lastm, false);
	lastm.show();

	hh = manage(new Gtk::HBox);
	hh->pack_start(button_area, true, false);
	button_area.show();
	vv->pack_start(* hh, false);
	hh->show();

	vv->pack_start(gs, false);
	gs.set_komi(komi);
	gs.show();

	vv->pack_start(gt, false);
	gt.show();

	b = manage(new Gtk::Button("Dup"));
	vv->pack_start(* b, false);
	b->clicked.connect(slot(this, & GoWin::duplicate));
	b->show();

	b = manage(new Gtk::Button("Close"));
	vv->pack_start(* b, false);
	b->clicked.connect(destroy.slot());
	b->show();

	// v->pack_start(gk, false);
	p->pack2(gk, true, false);
	gk.show();

	vv->show();
	h->show();
	v->show();
	p->show();
	p->set_position(1000);
	show();
	gk.line_enter.connect(talk.slot());
}

GoWin::~GoWin()
{
}

void GoWin::set_komi(float k)
{
	GoGame::set_komi(k);
	gs.set_komi(k);
}

void GoWin::n_move_change()
{
	dma.set_upper(n_moves());
	std::ostringstream s;
	s << n_moves();
	numm.set_text(s.str());
}

void GoWin::d_move_change()
{
	dma.set_value(dm);
	std::ostringstream s;
	if (dm == 0) s << "start";
	else switch (get_move(dm - 1).t) {
	case MOVE_PUT:
		s << (string) get_move(dm - 1).p;
		break;
	case MOVE_PASS:
		s << "pass";
		break;
	case MOVE_HANDICAP:
		s << "handicap";
		break;
	}
	lastm.set_text(s.str());
}


void GoWin::change_display()
{
	int to = int(dma.get_value());
	while (dm > to) back_display();
	while (dm < to) fore_display();
	d_move_change();
}



void GoWin::back_display()
{
	if (dm > 0) {
		dm --;
		Move mm = get_move(dm);
		if (mm.t == MOVE_PUT) {
			bd.mark(mm.p, GoBoard::NO_MARKING);
			if (mm.ko) bd.mark(mm.k[0], GoBoard::NO_MARKING);
			if (mm.k.size() > 0) {
				BoardStates s;
				if (mm.p == mm.k[0]) {
					s = TURN_STONE(mm.tn);
					gs.capture(OTHER_TURN(mm.tn),
						   - mm.k.size());
				} else {
					s = TURN_OTHER_STONE(mm.tn);
					gs.capture(mm.tn, - mm.k.size());
				}
				for (std::vector<Place>::const_iterator
				     i = mm.k.begin(); i != mm.k.end(); i ++)
					bd.put(* i, s);
			}
			bd.put(mm.p, NO_STONE);
		} else if (mm.t == MOVE_HANDICAP) {
			for (std::vector<Place>::const_iterator
				     i = mm.k.begin(); i != mm.k.end(); i ++)
				bd.put(* i, NO_STONE);
		}
		if (dm > 0) {
			mm = get_move(dm - 1);
			if (mm.t == MOVE_PUT)
				bd.mark(mm.p, GoBoard::LAST_MOVE);
			if (mm.ko) bd.mark(mm.k[0], GoBoard::KO_SPACE);
		}
	}
	bd.track(GoBoard::TRACK_NONE);
}

void GoWin::fore_display()
{
	if (dm < n_moves()) {
		Move mm;
		if (dm > 0) {
			mm = get_move(dm - 1);
			if (mm.t == MOVE_PUT)
				bd.mark(mm.p, GoBoard::NO_MARKING);
			if (mm.ko) bd.mark(mm.k[0], GoBoard::NO_MARKING);
		}
		mm = get_move(dm);
		if (mm.t == MOVE_PUT) {
			bd.put(mm.p, TURN_STONE(mm.tn));
			bd.mark(mm.p, GoBoard::LAST_MOVE);
			if (mm.k.size() > 0) {
				for (std::vector<Place>::const_iterator
					     i = mm.k.begin();
				     i != mm.k.end(); i ++)
					bd.put(* i, NO_STONE);
				if (mm.p == mm.k[0])
					gs.capture(OTHER_TURN(mm.tn),
						   mm.k.size());
				else gs.capture(mm.tn, mm.k.size());
				if (mm.ko) bd.mark(mm.k[0], GoBoard::KO_SPACE);
			}
		} else if (mm.t == MOVE_HANDICAP) {
			for (std::vector<Place>::const_iterator
				     i = mm.k.begin(); i != mm.k.end(); i ++)
				bd.put(* i, B_STONE);
		}
	        ++ dm;
	}
}

void GoWin::duplicate()
{
	GoWin * nw = manage(new GoWinPlay(z));
	nw->set_komi(komi);
	for (int i = 0; i < dm; i ++) nw->add_move(get_move(i));
	nw->n_move_change();
	while (nw->dm < dm) nw->fore_display();
	nw->d_move_change();
}

GoWinObserve::GoWinObserve(int bsiz) :
	GoGame(bsiz),
	GoWin(bsiz),
	status_bar("Observe Game")
{
	top_area.pack_start(status_bar);
	status_bar.show();
	magic = 0;
}

GoWinObserve::~GoWinObserve()
{
}

void GoWinObserve::add_move(const GoGame::Move & mm)
{
	if (mm.tn != turn) return;
	if (ms != STAGE_PLAYING) return;
	GoGame::add_move(mm);
	n_move_change();
	if (dm == (int) n_moves() - 1) {
		fore_display();
		d_move_change();
	}
	ostringstream s;
	s << "Move " << n_moves() << ": (" << (turn == B_TURN ? 'W' : 'B')
	  << ") ";
	switch (mm.t) {
	case MOVE_PUT:
		s << string(mm.p);
		break;
	case MOVE_HANDICAP:
		s << "Handicap " << mm.k.size();
		break;
	case MOVE_PASS:
		s << "Pass";
		break;
	}
	if (magic == 1) {
		gk.add_text(s.str(), GoTalk::TT_MOVES);
		magic = 0;
	}
}

void GoWinObserve::undo_move()
{
	// if (dm == 0) return;
	if (dm == n_moves()) back_display();
	d_move_change();
	move_undo();
	n_move_change();
	ostringstream s;
	s << (turn == B_TURN ? "White" : "Black") << " undid move "
	  << n_moves() + 1 << ".";
	gk.add_text(s.str(), GoTalk::TT_MOVES);
}

void GoWinObserve::set_id(const std::string & s)
{
	id_str = s;
}

const std::string & GoWinObserve::get_id()
{
	return id_str;
}

void GoWinObserve::set_white(const std::string & s)
{
	white = s;
}

const std::string & GoWinObserve::get_white()
{
	return white;
}

void GoWinObserve::set_black(const std::string & s)
{
	black = s;
}

const std::string & GoWinObserve::get_black()
{
	return black;
}

void GoWinObserve::set_status(const string & s)
{
	status_bar.set_text(s);
	gk.add_text(s, GoTalk::TT_MESSAGE);
}

void GoWinObserve::set_end()
{
	ms = STAGE_END;
	gt.stop();
	bd.track(GoBoard::TRACK_NONE);
}

void GoWinObserve::set_time(int wt, int wb, int bt, int bb)
{
	gt.set_time(wt, wb, bt, bb);
	gt.cont();
	magic = 1;
}

void GoWinObserve::set_time(const GameTime & t)
{
	gt.set_time(t);
	magic = 1;
}

void GoWinObserve::set_headline(const string & s)
{
	status_bar.set_text(s);
}

void GoWinObserve::set_board(const int * bb)
{
	for (int x = 0; x < z; x ++) for (int y = 0; y < z; y ++) {
		Place p(x, y);
		switch (bb[x + y * z]) {
		case 0:
			bd.put(p, B_STONE);
			break;
		case 1:
			bd.put(p, W_STONE);
			break;
		default:
			bd.put(p, NO_STONE);
		}
	}
}

void GoWinObserve::kibitz(const string & s)
{
	gk.add_text(s);
}

GoWinPlay::GoWinPlay(int sz) :
	GoGame(sz),
	GoWin(sz),
	button_undo("Undo"),
	pass_l("Pass")
{
	bd.hit.connect(slot(this, & GoWinPlay::hit_board));

	button_pass.add(pass_l);
	pass_l.show();
	button_area.pack_start(button_pass, false, false);
	button_pass.clicked.connect(slot(this, & GoWinPlay::hit_pass));
	button_pass.show();

	button_area.pack_start(button_undo, false, false);
	button_undo.clicked.connect(slot(this, & GoWinPlay::hit_undo));
	button_undo.show();

	bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
	gt.start();
	pass_count = 0;
}

GoWinPlay::~GoWinPlay()
{
}

void GoWinPlay::score_display()
{
	int wda = 0;
	int bda = 0;
	for (int x = 0; x < z; x ++) for (int y = 0; y < z ; y ++) {
		Place p(x, y);
		switch (terri(p)) {
		case B_TERRI:
			bd.mark(p, GoBoard::B_TERRI);
			bda ++;
			break;
		case W_TERRI:
			bd.mark(p, GoBoard::W_TERRI);
			wda ++;
			break;
		default:
			bd.mark(p, GoBoard::NO_MARKING);
		}
	}
	if (wda != wdame) gs.capture(W_TURN, wda - wdame);
	if (bda != bdame) gs.capture(B_TURN, bda - bdame);
	wdame = wda;
	bdame = bda;
}

void GoWinPlay::enter_score()
{
	ms = STAGE_SCORING;
	score_start();
	wdame = 0;
	bdame = 0;
	score();
	score_display();
	gt.stop();
	pass_l.set_text("Done");
	bd.track(GoBoard::TRACK_STONES);
}

void GoWinPlay::restore_board()
{
	restore();
	bd.clear();
	gs.reset();
	dm = 0;
	while (dm < n_moves()) fore_display();
	d_move_change();
	wdame = 0;
	bdame = 0;
	score();
	score_display();
	// bd.track(GoBoard::TRACK_STONES);
}

void GoWinPlay::remove_group(const Place & p)
{
	vector<Place> l;
	get_group(p, l);
	rid_group(p);
	if (l.size() > 0)
		gs.capture(STONE_OTHER_TURN(bd.get(p)), l.size());
	for (vector<Place>::const_iterator i = l.begin();
	i != l.end(); i ++)
		bd.put(* i, NO_STONE);
	score();
	score_display();
	bd.track(GoBoard::TRACK_STONES);
}

void GoWinPlay::hit_board(const Place & p)
{
	bd.track(GoBoard::TRACK_NONE);
	if (dm < n_moves()) return;
	if (ms == STAGE_PLAYING) {
		gt.switching();
		move_put(p);
		n_move_change();
		fore_display();
		d_move_change();
		bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
		pass_count = 0;
	} else if (ms == STAGE_SCORING) {
		remove_group(p);
	}
}

void GoWinPlay::hit_pass()
{
	bd.track(GoBoard::TRACK_NONE);
	if (dm < n_moves()) return;
	if (ms == STAGE_PLAYING) {
		gt.switching();
		move_pass();
		n_move_change();
		fore_display();
		d_move_change();
		bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
		if (++ pass_count >= 3) {
			pass_count = 0;
			enter_score();
		}
	} else if (ms == STAGE_SCORING) {
		// done
		ms = STAGE_END;
	}
}

void GoWinPlay::hit_undo()
{
	bd.track(GoBoard::TRACK_NONE);

	if (dm < n_moves()) return;
	if (ms == STAGE_PLAYING) {
		gt.switching();
		back_display();
		d_move_change();
		move_undo();
		n_move_change();
		bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
		pass_count = 0;
	} else if (ms == STAGE_SCORING) {
		restore_board();
		bd.track(GoBoard::TRACK_STONES);
	}
}

void GoWinPlay::fore_display()
{
	bd.track(GoBoard::TRACK_NONE);
	GoWin::fore_display();
	if (dm == n_moves()) {
		if (ms == STAGE_PLAYING) {
			bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
		} else if (ms == STAGE_SCORING) {
			bd.track(GoBoard::TRACK_STONES);
		}
	}
}

void GoWinPlay::back_display()
{
	bd.track(GoBoard::TRACK_NONE);
	GoWin::back_display();
}

GoWinMatch::GoWinMatch(int sz, bool pw, bool pb) :
	GoGame(sz),
	GoWin(sz),
	GoWinObserve(sz),
	GoWinPlay(sz)
{
	play_turn[W_TURN] = pw;
	play_turn[B_TURN] = pb;
	if (play_turn[turn])
		bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
	else {
		bd.track(GoBoard::TRACK_NONE);
		button_pass.set_sensitive(false);
		button_undo.set_sensitive(false);
	}
}

GoWinMatch::~GoWinMatch()
{
}

void GoWinMatch::add_move(const Move & m)
{
	GoWinObserve::add_move(m);
	if (ms == STAGE_PLAYING) {
		if (play_turn[turn]) {
			bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
			button_pass.set_sensitive(true);
			button_undo.set_sensitive(true);
		} else {
			bd.track(GoBoard::TRACK_NONE);
			button_pass.set_sensitive(false);
			button_undo.set_sensitive(false);	
		}
	} else if (ms == STAGE_SCORING) { // impossible??
		bd.track(GoBoard::TRACK_STONES);
		button_pass.set_sensitive(true);
		button_undo.set_sensitive(true);
	}
}

void GoWinMatch::undo_move()
{
	GoWinObserve::undo_move();
	if (ms == STAGE_PLAYING) {
		if (play_turn[turn]) {
			bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
			button_pass.set_sensitive(true);
			button_undo.set_sensitive(true);
		} else {
			bd.track(GoBoard::TRACK_NONE);
			button_pass.set_sensitive(false);
			button_undo.set_sensitive(false);
		}
	} else if (ms == STAGE_SCORING) { // impossible??
		bd.track(GoBoard::TRACK_STONES);
		button_pass.set_sensitive(true);
		button_undo.set_sensitive(true);
	}
}

void GoWinMatch::set_end()
{
	GoWinObserve::set_end();
	bd.track(GoBoard::TRACK_NONE);
	button_pass.set_sensitive(false);
	button_undo.set_sensitive(false);
}

void GoWinMatch::hit_board(const Place & p)
{
	bd.track(GoBoard::TRACK_NONE);
	if (dm < (int) n_moves()) return;
	board_hit.emit(p);
	if (ms == STAGE_PLAYING)
		bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
	else if (ms == STAGE_SCORING) bd.track(GoBoard::TRACK_STONES);
}

void GoWinMatch::hit_pass()
{
	bd.track(GoBoard::TRACK_NONE);
	if (dm < (int) n_moves()) return;
	if (ms == STAGE_PLAYING) {
		pass_hit.emit();
		bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
	} else if (ms == STAGE_SCORING) {	
		done_hit.emit();
		bd.track(GoBoard::TRACK_STONES);
	}
}

void GoWinMatch::hit_undo()
{
	bd.track(GoBoard::TRACK_NONE);
	if (dm < (int) n_moves()) return;
	undo_hit.emit();
	if (ms == STAGE_PLAYING)
		bd.track(GoBoard::TRACK_SPACE, TURN_STONE(turn));
	else if (ms == STAGE_SCORING) bd.track(GoBoard::TRACK_STONES);
}
